;;
;;  Parser for the grammar defined in RFC 5322, "Internet Message Format".
;;
;;  Based on the Haskell Rfc2822 module by Peter Simons.
;;
;;   Copyright 2009 Ivan Raikov and the Okinawa Institute of Science and Technology.
;;
;;
;;   This program is free software: you can redistribute it and/or
;;   modify it under the terms of the GNU General Public License as
;;   published by the Free Software Foundation, either version 3 of the
;;   License, or (at your option) any later version.
;;
;;   This program is distributed in the hope that it will be useful, but
;;   WITHOUT ANY WARRANTY; without even the implied warranty of
;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;   General Public License for more details.
;;
;;   A full copy of the GPL license can be found at
;;   <http://www.gnu.org/licenses/>.

(module internet-message

	(comment fields body message )

	(import scheme chicken data-structures srfi-1 srfi-14)

	(require-library srfi-1 abnf abnf-consumers)
	(import (prefix abnf abnf:) 
		(prefix abnf-consumers abnf:) 
		)

(define (char-list-titlecase x)
  (if (null? x) x (cons (char-upcase (car x)) (map char-downcase (cdr x)))))

;; construct symbols from consumed chars
(define consumed-chars->tsymbol
  (abnf:consumed-chars->list 
   (compose string->symbol 
	    list->string 
	    char-list-titlecase 
	    abnf:trim-ws-char-list)))

;; shortcut for (abnf:bind consumed-chars->tsymbol (abnf:longest ... ))
(define-syntax bind-consumed->tsymbol
  (syntax-rules () 
    ((_ p)    (abnf:bind consumed-chars->tsymbol (abnf:longest p)))
    ))

(define consumed-objects-lift-any
  (abnf:consumed-objects-lift
   (abnf:consumed-objects identity)))

;; Construct a parser for a message header line from the header's name
;; and a parser for the body.

(define (header s p)  
  (let ((ss (->string s)))
    (abnf:bind (consumed-objects-lift-any)
     (abnf:concatenation
      (bind-consumed->tsymbol (abnf:lit ss))
      (abnf:drop-consumed (abnf:char #\:))
      p
      (abnf:drop-consumed abnf:crlf)
      ))))


;; Primitive parsers (section 3.2.1)


;; Match any US-ASCII character except for nul \r \n

(define text (abnf:set (char-set-difference
			char-set:ascii 
			(char-set (integer->char 0) 
				  (integer->char 10) 
				  (integer->char 13) ))))


;; Folding white space and comments (section 3.2.3)

(define fws
  (abnf:concatenation
   (abnf:optional-sequence 
    (abnf:concatenation
     (abnf:repetition abnf:wsp)
     (abnf:drop-consumed abnf:crlf)))
   (abnf:repetition1 abnf:wsp)))

(define (between-fws-drop p)
  (abnf:concatenation
   (abnf:drop-consumed (abnf:optional-sequence fws)) p 
   (abnf:drop-consumed (abnf:optional-sequence fws))))
			     

;; helper macro for mutually-recursive parser definitions

(define-syntax vac
  (syntax-rules ()
    ((_ fn) (lambda args (apply fn args)))))


;; Match any non-whitespace, non-control character except for ( ) and \

(define ctext  (abnf:set (char-set-difference char-set:graphic (char-set #\( #\) #\\))))

;; Match comments. That is any combination of ctext, quoted pairs, and
;; fws between brackets. Comments may nest.

(define ccontent (vac (abnf:alternatives ctext abnf:quoted-pair comment)))

(define comment (abnf:concatenation 
		 (abnf:char #\( )
		 (abnf:longest
		  (abnf:repetition
		   (abnf:concatenation
		    (abnf:optional-sequence fws)
		    ccontent
		    )))
		 (abnf:optional-sequence fws)
		 (abnf:char #\))
		 ))

;; Match any combination of fws and comments

(define cfws (abnf:alternatives
	      (abnf:concatenation
	       (abnf:repetition1
		(abnf:concatenation
		 (abnf:optional-sequence fws)
		 (abnf:drop-consumed comment)))
	       (abnf:optional-sequence fws))
	      fws))
		 

;;  A combinator for sequences (optional cfws) p (optional cfws)

(define (between-cfws p)
  (abnf:concatenation
   (abnf:optional-sequence cfws) p 
   (abnf:optional-sequence cfws) ))
			     
(define (between-cfws-drop p)
  (abnf:concatenation
   (abnf:drop-consumed (abnf:optional-sequence cfws)) p 
   (abnf:drop-consumed (abnf:optional-sequence cfws) )))
			     

;; Atom (section 3.2.4)

;; Match any US-ASCII character except for control characters,
;; specials, or space. atom and dot-atom are made up of this.

(define atext (abnf:alternatives
	       abnf:alpha
	       abnf:decimal
	       (abnf:set-from-string "!#$%&'*+-/=?^_`{|}~")))

;; Match one or more atext characters and skip any preceeding or
;; trailing cfws

(define atom (abnf:bind-consumed->string (between-cfws (abnf:repetition1 atext))))

;; Match two or more atext elements interspersed by dots.

(define dot-atom-text (abnf:concatenation
		       (abnf:repetition1 atext)
		       (abnf:repetition 
			(abnf:concatenation
			 (abnf:char #\.)
			 (abnf:repetition1 atext) 
			 ))))

;; Match dot-atom-text and skip any preceeding or trailing cfws.

(define dot-atom (abnf:bind-consumed->string (between-cfws dot-atom-text)))

;; Quoted strings (section 3.2.4)

;;; Match any non-whitespace, non-control US-ASCII character except
;;; for \ and "

(define char-set:quoted (char-set-difference char-set:printing (char-set #\\ #\")))
(define qtext (abnf:set char-set:quoted))


;; Match either qtext or quoted-pair.

(define qcontent (abnf:repetition1 
		  (abnf:alternatives
		   qtext abnf:quoted-pair)))

;; Match any number of qcontent between double quotes. 

(define quoted-string
  (abnf:bind-consumed->string
   (between-cfws 
    (abnf:concatenation
     (abnf:drop-consumed abnf:dquote)
     (abnf:repetition
      (abnf:concatenation
       (abnf:optional-sequence fws)
       qcontent))
     (abnf:optional-sequence fws)
     (abnf:drop-consumed abnf:dquote)))))

;; Miscellaneous tokens (section 3.2.5)

;;; Match either atom or quoted-string

(define word (abnf:alternatives atom quoted-string))

;; Match either one or more word elements 

(define phrase 
  (abnf:bind-consumed-strings->list (abnf:repetition1 word)))


;; Match any number of utext tokens.
;;
;; Unstructured text is used in free text fields such as subject.

(define unstructured 
  (abnf:bind-consumed->string
   (abnf:concatenation
    (abnf:repetition 
     (abnf:concatenation
      (abnf:optional-sequence fws)
      abnf:vchar))
    (abnf:repetition abnf:wsp))))

;; Date and Time Specification (section 3.3)

;; Parses a date and time specification of the form
;;
;;   Thu, 19 Dec 2002 20:35:46 +0200
;;
;; where the weekday specification (Thu) is optional. The parser
;; This parser will not perform any consistency checking.
;; It will accept
;;
;;    40 Apr 2002 13:12 +0100
;;
;;  as a perfectly valid date.

			     
;; Match the abbreviated weekday names

(define day-name 
  (abnf:alternatives
   (abnf:lit "Mon")
   (abnf:lit "Tue")
   (abnf:lit "Wed")
   (abnf:lit "Thu")
   (abnf:lit "Fri")
   (abnf:lit "Sat")
   (abnf:lit "Sun")))

;; Match a day-name, optionally wrapped in folding whitespace

(define day-of-week 
  (abnf:bind-consumed-strings->list 
   'day-of-week 
   (between-fws-drop 
    (abnf:bind-consumed->string day-name))))


;; Match a four digit decimal number

(define year 
  (between-fws-drop
   (abnf:bind-consumed->string (abnf:repetition-n 4 abnf:decimal))))

;; Match the abbreviated month names


(define month-name (abnf:alternatives
		    (abnf:lit "Jan")
		    (abnf:lit "Feb")
		    (abnf:lit "Mar")
		    (abnf:lit "Apr")
		    (abnf:lit "May")
		    (abnf:lit "Jun")
		    (abnf:lit "Jul")
		    (abnf:lit "Aug")
		    (abnf:lit "Sep")
		    (abnf:lit "Oct")
		    (abnf:lit "Nov")
		    (abnf:lit "Dec")))

;; Match a month-name, optionally wrapped in folding whitespace

(define month (between-fws-drop (abnf:bind-consumed->string month-name)))


;; Match a one or two digit number

(define day (abnf:concatenation
	     (abnf:drop-consumed (abnf:optional-sequence fws))
	     (abnf:alternatives 
	      (abnf:bind-consumed->string (abnf:variable-repetition 1 2 abnf:decimal))
	      (abnf:drop-consumed fws))))

;; Match a date of the form dd:mm:yyyy
(define date 
  (abnf:bind-consumed-strings->list 'date 
    (abnf:concatenation day month year)))

;; Match a two-digit number 

(define hour      (abnf:bind-consumed->string (abnf:repetition-n 2 abnf:decimal)))
(define minute    (abnf:bind-consumed->string (abnf:repetition-n 2 abnf:decimal)))
(define isecond   (abnf:bind-consumed->string (abnf:repetition-n 2 abnf:decimal)))

;; Match a time-of-day specification of hh:mm or hh:mm:ss.

(define time-of-day (abnf:concatenation
		     hour (abnf:drop-consumed (abnf:char #\:))
		     minute (abnf:optional-sequence 
			     (abnf:concatenation (abnf:drop-consumed (abnf:char #\:))
						 isecond))))

;; Match a timezone specification of the form
;; +hhmm or -hhmm 

(define zone (abnf:concatenation 
	      (abnf:drop-consumed fws)
	      (abnf:bind-consumed->string (abnf:alternatives (abnf:char #\-) (abnf:char #\+)))
	      hour minute))

;; Match a time-of-day specification followed by a zone.

(define itime 
  (abnf:bind-consumed-strings->list 'time 
    (abnf:concatenation time-of-day zone)))

(define date-time (abnf:concatenation
		   (abnf:optional-sequence
		    (abnf:concatenation
		     day-of-week
		     (abnf:drop-consumed (abnf:char #\,))))
		   date
		   itime
		   (abnf:drop-consumed (abnf:optional-sequence cfws))))


;; Address Specification (section 3.4)


;; Parse and return a "local part" of an addr-spec. That is either
;; a dot-atom or a quoted-string.

(define local-part (abnf:alternatives dot-atom quoted-string))


;; Parse and return any ASCII characters except [ ] and \

(define dtext  (abnf:set (char-set-difference char-set:printing (char-set #\[ #\] #\\))))


;; Parse a domain literal. That is a [ character, followed by any
;; amount of dcontent, followed by a terminating ] character.

(define domain-literal (between-cfws
			 (abnf:concatenation
			  (abnf:drop-consumed (abnf:char #\[))
			  (abnf:bind-consumed->string 
			   (abnf:repetition 
			    (abnf:concatenation
			     (abnf:drop-consumed (abnf:optional-sequence fws))
			     dtext)))
			  (abnf:drop-consumed (abnf:optional-sequence fws))
			  (abnf:drop-consumed (abnf:char #\]))
			  )))

;; Parse and return a domain part of an addr-spec. That is either
;; a dot-atom or a domain-literal.

(define domain  (abnf:alternatives dot-atom domain-literal))


;; Addr-spec specification (section 3.4.1)

;; Parse an address specification. That is, a local-part, followed
;; by an \ character, followed by a domain. 

(define addr-spec
  (abnf:concatenation
   (abnf:bind-consumed-strings->list 'local-part local-part)
   (abnf:drop-consumed (abnf:char #\@))
   (abnf:bind-consumed-strings->list 'domain domain)))

;; Parse an angle-addr

(define angle-addr
  (between-cfws-drop
   (abnf:concatenation
    (abnf:drop-consumed (abnf:char #\<))
    addr-spec
    (abnf:drop-consumed (abnf:char #\>))
    )))


;; Parse and return a phrase.
(define display-name  (abnf:bind-consumed-pairs->list 'display-name phrase))

;; Match an angle-addr, optionally prefaced with a display-name

(define name-addr
  (abnf:concatenation
   (abnf:optional-sequence display-name)
   angle-addr))

;; Match a name-addr or an addr-spec and return the address.

(define mailbox 
  (abnf:bind-consumed-pairs->list 'mailbox 
   (abnf:alternatives name-addr addr-spec)))

;; Parse a list of mailbox addresses, every two addresses being
;; separated by a comma, and return the list of found address(es).

(define mailbox-list
  (abnf:bind-consumed-pairs->list 'mailbox-list
   (abnf:concatenation
    mailbox
    (abnf:repetition
     (abnf:concatenation
      (abnf:drop-consumed (abnf:char #\,))
      mailbox)))))


;; Parse a group of addresses. That is, a display-name, followed
;; by a colon, optionally followed by a mailbox-list, followed by a
;; semicolon. The found address(es) are returned - what may be none.
;; Here is an example:
;;
;;    my group: user1@example.org, user2@example.org;

(define group (vac (abnf:bind-consumed-pairs->list 'group 
		   (abnf:concatenation
		    display-name
		    (abnf:drop-consumed (abnf:char #\:))
		    (abnf:optional-sequence group-list)
		    (abnf:drop-consumed (abnf:char #\;))
		    (abnf:drop-consumed (abnf:optional-sequence cfws))))))
  
(define group-list (abnf:alternatives 
		    mailbox-list
		    (abnf:drop-consumed cfws)))

;; Matches a single mailbox or an address group 

(define address (abnf:alternatives mailbox group))

;; Parse a list of address addresses, every two addresses being
;; separated by a comma, and return the list of found address(es).

(define address-list (abnf:concatenation
		      address
		      (abnf:repetition
		       (abnf:concatenation
			(abnf:drop-consumed (abnf:char #\,))
			address))))

;;  Overall message syntax (section 3.5)
			     
;; This parser will return a message body as specified by the RFC;
;; that is basically any number of text characters, which may be
;; divided into separate lines by crlf.

(define body (abnf:repetition
	      (abnf:concatenation
	       (abnf:repetition 
		 (abnf:concatenation
		  (abnf:bind-consumed->string
		   (abnf:repetition text))
		  (abnf:drop-consumed 
		   (abnf:repetition abnf:crlf))))
	       (abnf:bind-consumed->string
		(abnf:repetition text)))))

;; Field definitions (section 3.6)

;; The origination date field (section 3.6.1)

;; Parse a Date: header and return the date as a list
;; (year month dom hour min sec tz dow)

(define orig-date  (header "Date" date-time))

;; Originator fields (section 3.6.2)

;; Parse a From: header and return the mailbox-list address(es)
;; contained in it.

(define from      (header "From" mailbox-list))

;; Parse a Sender: header and return the mailbox address contained in
;; it.

(define sender    (header "Sender" mailbox))

;; Parse a Reply-To: header and return the address-list address(es)
;; contained in it.

(define reply-to  (header "Reply-To" address-list))

;; Destination address fields (section 3.6.3)

;; Parse a To: header and return the address-list address(es)
;; contained in it.

(define to        (header "To" address-list))

;; Parse a Cc: header and return the address-list address(es)
;; contained in it.

(define cc        (header "Cc" address-list))

;; Parse a Bcc: header and return the address-list address(es)
;; contained in it.

(define bcc       (header "Bcc" (abnf:optional-sequence 
				 (abnf:alternatives
				  address-list
				  (abnf:drop-consumed cfws)))))

;; Identification fields (section 3.6.4)

;; Parse one or more occurences of dtext or quoted-pair and return the
;; concatenated string. This makes up the id-right of a msg-id.

(define no-fold-literal  (abnf:concatenation
			  (abnf:drop-consumed (abnf:char #\[))
			  (abnf:repetition dtext)
			  (abnf:drop-consumed (abnf:char #\]))))

;; Parse a left ID part of a msg-id. This is almost identical to
;; the local-part of an e-mail address, but with stricter rules
;; about folding and whitespace.

(define id-left   dot-atom-text )

;; Parse a right ID part of a msg-id. This is almost identical to the
;; domain of an e-mail address, but with stricter rules about folding
;; and whitespace.

(define id-right  (abnf:alternatives dot-atom-text no-fold-literal))

;; Parse a message ID and return it. A message ID is almost identical
;; to an angle-addr, but with stricter rules about folding and
;; whitespace.

(define msg-id
  (abnf:bind-consumed-strings->list 'message-id
    (between-cfws-drop
     (abnf:concatenation
      (abnf:drop-consumed (abnf:char #\<))
      (abnf:bind-consumed->string id-left)
      (abnf:drop-consumed (abnf:char #\@))
      (abnf:bind-consumed->string id-right)
      (abnf:drop-consumed (abnf:char #\>))
      ))))


;; Parse a In-Reply-To header and return the list of msg-id's
;; contained in it.

(define in-reply-to (header "In-Reply-To" (abnf:repetition1 msg-id)))

;; Parse a References: header and return the list of msg-id's
;; contained in it.

(define references  (header "References"  (abnf:repetition1 msg-id)))

;; Parse a Message-Id: header and return the msg-id contained
;; in it.

(define message-id (header "Message-ID" msg-id))
					   
;; Informational fields (section 3.6.5)

;; Parse a Subject: header and return its contents verbatim.

(define subject (header "Subject"  unstructured))

;; Parse a Comments: header and return its contents verbatim.


(define comments (header "Comments" unstructured))

;; Parse a Keywords: header and return the list of phrases
;; found. Please note that each phrase is again a list of atoms, as
;; returned by the phrase parser.

(define keywords  (header "Keywords"
			   (abnf:concatenation
			    phrase
			    (abnf:repetition
			     (abnf:concatenation
			      (abnf:drop-consumed (abnf:char #\,))
			      phrase)))))

;; Resent fields (section 3.6.6)

;; Parse a Resent-Date: header and return the date it contains as
;; CalendarTime

(define resent-date (header "Resent-Date" date-time))

;; Parse a Resent-From: header and return the mailbox-list address(es)
;; contained in it.

(define resent-from  (header "Resent-From" mailbox-list))

;; Parse a Resent-Sender: header and return the mailbox-list
;; address(es) contained in it.

(define resent-sender (header "Resent-Sender" mailbox))

;; Parse a Resent-To header and return the mailbox address contained
;; in it.

(define resent-to  (header "Resent-To" address-list))

;; Parse a Resent-Cc header and return the address-list address(es)
;; contained in it.

(define resent-cc (header "Resent-Cc" address-list))

;; Parse a Resent-Bcc: header and return the address-list
;; address(es) contained in it. (This list may be empty.)

(define resent-bcc   (header "Resent-Bcc"
			     (abnf:alternatives 
			      address-list
			      (abnf:drop-consumed (abnf:optional-sequence cfws)))))


;; Parse a Resent-Message-ID: header and return the msg-id contained
;; in it.

(define resent-msg-id  (header "Resent-Message-ID" msg-id))


;; Parse a Resent-Reply-To: header and return the address-list
;; contained in it.

(define resent-reply-to  (header "Resent-Reply-To" address-list))


;; Trace fields (section 3.6.7)

			 
(define path    (abnf:alternatives 
		 angle-addr
		 (between-cfws-drop
		 (abnf:concatenation
		   (abnf:drop-consumed (abnf:char #\<))
		   (abnf:drop-consumed (abnf:optional-sequence cfws))
		   (abnf:drop-consumed (abnf:char #\>))))))

(define return-path  (header "Return-Path" path))

(define received-token 
  (abnf:bind-consumed-strings->list
   'received-token
   (abnf:alternatives domain angle-addr addr-spec word)))

(define received  (header "Received"  
			   (abnf:concatenation
			    (abnf:repetition received-token)
			    (abnf:drop-consumed (abnf:char #\;))
			    date-time)))


;; Optional fields (section 3.6.8)

;; Match and return any ASCII character except for control
;; characters, whitespace, and :

(define ftext  (abnf:set (char-set-difference char-set:graphic
					      (char-set #\:))))


;; Parse and return an arbitrary header field name. That is one or
;; more ftext characters.

(define field-name (bind-consumed->tsymbol (abnf:repetition1 ftext)))

;; Parse an arbitrary header field and return a tuple containing the
;; field-name and unstructured text of the header. The name will not
;; contain the terminating colon.

(define optional-field  
  (abnf:bind (consumed-objects-lift-any)
	     (abnf:concatenation
	      field-name
	      (abnf:drop-consumed (abnf:char #\:))
	      unstructured
	      (abnf:drop-consumed abnf:crlf))))
  
;; This parser will parse an arbitrary number of header fields as
;; defined in this RFC. For each field, an appropriate 'Field' value
;; is created, all of them making up the 'Field' list that this parser
;; returns.

;; Fields that contain syntax errors fall back to the catch-all
;; optional-field. Thus, this parser will hardly ever return a syntax
;; error -- what conforms with the idea that any message that can
;; possibly be accepted /should/ be.

(define fields
  (abnf:longest
   (abnf:repetition
    (abnf:alternatives
     from
     sender
     return-path
     reply-to
     to
     cc
     bcc
     message-id
     in-reply-to
     references
     subject
     comments
     keywords
     orig-date
     resent-date
     resent-from
     resent-sender
     resent-to
     resent-cc
     resent-bcc
     resent-msg-id
     resent-reply-to
     received
     optional-field))))



;; Parse a complete message as defined by the RFC and returns
;;  the separate header fields and the message body. 

(define message
  (abnf:bind-consumed-pairs->list 
   'message
   (abnf:concatenation 
    (abnf:bind-consumed-pairs->list 'fields fields)
    (abnf:optional-sequence
     (abnf:concatenation 
      (abnf:drop-consumed abnf:crlf)
      (abnf:bind-consumed-strings->list 'body body))))))


)
